/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import static java.util.Collections.*;

import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.Spliterators;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormData.Representations;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
class XtaService {

	static final String XDOMEA_0201_MESSAGE_TYPE = "Geschaeftsgang.Geschaeftsgang.0201";

	static final String FIM_MESSAGE_TYPE_PREFIX = "fim.";

	@Autowired
	private XtaRemoteService remoteService;
	@Autowired
	private XtaMessageMapper mapper;
	@Autowired
	private VorgangNummerSupplier vorgangNummerSupplier;
	@Autowired
	private XdomeaMessageDataMapper xdomeaMessageDataMapper;
	@Autowired
	private XtaIncomingFilesMapper xtaIncomingFilesMapper;
	@Autowired
	private XtaCurrentIdentifierService service;

	public Stream<FormData> getMessages() {
		return createXtaMessageStream()
				.filter(this::isSupportedMessageType)
				.map(this::getFormDataIfNoRuntimeException)
				.flatMap(Optional::stream);
	}

	Stream<XtaMessageMetaData> createXtaMessageStream() {
		return service.getIdentifiers().stream()
				.flatMap(this::createXtaMessageStreamForIdentifier);
	}

	private Stream<XtaMessageMetaData> createXtaMessageStreamForIdentifier(String identifier) {
		service.setCurrentIdentifier(identifier);
		var iterator = new XtaMessageMetadataRemoteIterator(remoteService);
		return StreamSupport.stream(Spliterators.spliteratorUnknownSize(iterator, 0), false);
	}

	boolean isSupportedMessageType(XtaMessageMetaData metaData) {
		if (isXDomeaMessageType(metaData.getMessageType()) || isFimMessageType(metaData.getMessageType())) {
			return true;
		}
		LOG.warn("Ignoring XTA-Message of type '{}'.", metaData.getMessageType());
		return false;
	}

	Optional<FormData> getFormDataIfNoRuntimeException(@NonNull XtaMessageMetaData metaData) {
		try {
			return Optional.of(getFormData(metaData));
		} catch (RuntimeException exception) {
			LOG.error("Failed to process xta message (id: %s)".formatted(metaData.getMessageId()), exception);
			return Optional.empty();
		}
	}

	FormData getFormData(@NonNull XtaMessageMetaData metaData) {
		var msg = remoteService.getMessage(metaData.getMessageId());
		var incomingFiles = xtaIncomingFilesMapper.toIncomingFiles(msg.getMessageFiles());
		var representationsAttachmentsPair = getRepresentationsAttachmentsPair(metaData, incomingFiles);

		var formData = mapper.toFormData(representationsAttachmentsPair, metaData, vorgangNummerSupplier);
		formData = addAttachments(msg, formData);

		return addRepresentations(formData, msg);
	}

	FormData addRepresentations(FormData formData, XtaMessage msg) {
		return formData.toBuilder().control(
				formData.getControl().toBuilder()
						.representations(Optional.of(buildRepresentations(formData.getControl().getRepresentations(), msg)))
						.build())
				.build();
	}

	private Representations buildRepresentations(Optional<Representations> base, XtaMessage msg) {
		return base.map(Representations::toBuilder).orElseGet(Representations::builder)
				.primaryFormDataRepresentation(msg.getPrimaryFormDataMessage())
				.build();
	}

	FormData addAttachments(XtaMessage msg, FormData inFormData) {
		var attachments = buildAttachmentsInFiles(msg.getAttachments());
		if (CollectionUtils.isNotEmpty(attachments)) {
			return inFormData.toBuilder()
					.attachment(IncomingFileGroup.builder().name("sonstige").files(attachments).build())
					.numberOfAttachments(attachments.size())
					.build();
		}
		return inFormData;
	}

	private List<IncomingFile> buildAttachmentsInFiles(Collection<XtaFile> attachmentFiles) {
		return attachmentFiles.stream().map(xtaIncomingFilesMapper::toIncomingFile).toList();
	}

	RepresentationsAttachmentsPair getRepresentationsAttachmentsPair(XtaMessageMetaData metaData, List<IncomingFile> incomingFiles) {
		if (isXDomeaMessageType(metaData.getMessageType())) {
			return xdomeaMessageDataMapper.mapIncomingFilesToRepresentationsAttachmentsPair(incomingFiles);
		}
		if (isFimMessageType(metaData.getMessageType())) {
			return mapIncomingFilesToRepresentations(incomingFiles);
		}

		throw new TechnicalException("Unexpected XTA message type: %s".formatted(metaData.getMessageType()));
	}

	private boolean isXDomeaMessageType(String messageType) {
		return StringUtils.equals(messageType, XDOMEA_0201_MESSAGE_TYPE);
	}

	private boolean isFimMessageType(String messageType) {
		return StringUtils.startsWith(messageType, FIM_MESSAGE_TYPE_PREFIX);
	}

	RepresentationsAttachmentsPair mapIncomingFilesToRepresentations(List<IncomingFile> incomingFiles) {
		return RepresentationsAttachmentsPair.builder()
				.representations(incomingFiles)
				.attachments(emptyList())
				.build();
	}

	public void acknowledgeReceive(@NonNull XtaMessageId messageId) {
		remoteService.close(messageId);
	}
}
